import math
import numpy as np


def transform_coor(unit_vectors, xyz):
    '''
    Transform the coordinates in new system back to its coordinates in old system

    Parameters
    ----------
    unit_vectors : list of array
        unit_vectors of the new system in old system
    xyz : array
        coordinates in new system

    Returns
    -------
    xyz_old : array
        coordinates in old system
    '''
    R = np.array(unit_vectors).T
    b = np.matmul(R, xyz[:, np.newaxis])
    return b[:, 0]


def grow_particle(pos1, pos2, bond, angle):
    '''
    Generate a random coordinates based on the coordinates of two parents, bond and angle

    https://scicomp.stackexchange.com/questions/27965/rotate-a-vector-by-a-randomly-oriented-angle
    @tparker

    Parameters
    ----------
    pos1 : np.ndarray
        position of far parent
    pos2 : np.ndarray
        position of near parent
    bond : float
        bond between near parent and new particle
    angle : float
        angle between far parent, near parent and new particle

    Returns
    -------
    pos3 : array
        position of new particle
    '''
    delta = pos2 - pos1
    v = delta / np.sqrt(np.dot(delta, delta))

    a = np.array([1, 0, 0])
    if np.dot(a, v) > 0.99:
        a = np.array([0, 1, 0])
    xb = np.cross(a, v)
    xb = xb / np.sqrt(np.dot(xb, xb))
    yb = np.cross(v, xb)

    phi = np.random.random() * 2 * np.pi
    theta = np.pi - angle
    w = np.sin(theta) * np.cos(phi) * xb + np.sin(theta) * np.sin(phi) * yb + np.cos(theta) * v

    return pos2 + w * bond


def relocate_hydrogen(hydrogen):
    '''
    Update the position of a hydrogen atom according to the parent atom and other neighbors of parent

    Parameters
    ----------
    hydrogen : Atom
    '''
    vector = np.array([0., 0., 0.])
    parent = hydrogen.bond_partners[0]
    for atom in parent.bond_partners:
        if atom is not hydrogen:
            vector += atom.position - parent.position
    hydrogen.position = parent.position - vector / np.sqrt(np.dot(vector, vector)) * 0.1  # bXH = 0.1 nm


def periodic_distance(pos1, pos2, box, distance_max=None):
    '''
    Calculate the distance between two points under periodic boundary condition
    
    Parameters
    ----------
    pos1 : np.ndarray
    pos2 : np.ndarray
    box : np.ndarray of shape (3,)
        Lengths of rectangular periodic box. Triclinic box is not supported yet.
    distance_max : float, optional
        The maximum distance to be considered. Will return None if distance larger than it

    Returns
    -------
    distance : float or None
        May return None if distance is apparently larger than `distance_max`
    '''
    delta = pos2 - pos1
    abs_delta = np.abs(delta)
    if distance_max is not None and any((abs_delta > distance_max) & (abs_delta < box - distance_max)):
        return None

    ### elements of delta will be transformed to (-0.5, 0.5]
    delta -= np.ceil(delta / box - 0.5) * box

    ### elements of delta will be transformed to [-0.5, 0.5)
    # delta -= np.floor(delta / box + 0.5) * box

    return math.sqrt(np.dot(delta, delta))


def periodic_angle(pos1, pos2, pos3, box):
    '''
    Calculate the angle between three points under periodic boundary condition

    Parameters
    ----------
    pos1 : np.ndarray
    pos2 : np.ndarray
    pos3 : np.ndarray
    box : np.ndarray of shape (3,)
        Lengths of rectangular periodic box. Triclinic box is not supported yet.

    Returns
    -------
    angle : float
    '''
    vec1 = pos1 - pos2
    vec2 = pos3 - pos2

    ### elements of delta will be transformed to (-0.5, 0.5]
    vec1 -= np.ceil(vec1 / box - 0.5) * box
    vec2 -= np.ceil(vec2 / box - 0.5) * box

    cos = vec1.dot(vec2) / np.sqrt(vec1.dot(vec1) * vec2.dot(vec2))
    return float(np.arccos(np.clip(cos, -1, 1)))


def periodic_dihedral(pos1, pos2, pos3, pos4, box):
    '''
    Calculate the dihedral between four points under periodic boundary condition

    Parameters
    ----------
    pos1 : np.ndarray
    pos2 : np.ndarray
    pos3 : np.ndarray
    pos4 : np.ndarray
    box : np.ndarray of shape (3,)
        Lengths of rectangular periodic box. Triclinic box is not supported yet.

    Returns
    -------
    angle : float
    '''
    vec1 = pos2 - pos1
    vec2 = pos3 - pos2
    vec3 = pos4 - pos3

    ### elements of delta will be transformed to (-0.5, 0.5]
    vec1 -= np.ceil(vec1 / box - 0.5) * box
    vec2 -= np.ceil(vec2 / box - 0.5) * box
    vec3 -= np.ceil(vec3 / box - 0.5) * box

    n1 = np.cross(vec1, vec2)
    n2 = np.cross(vec2, vec3)
    cos = n1.dot(n2) / np.sqrt(n1.dot(n1) * n2.dot(n2))
    value = float(np.arccos(np.clip(cos, -1, 1)))
    sign = 1 if vec1.dot(n2) >= 0 else -1
    return sign * value


def rotate_points(positions, reference, target):
    '''
    Rotate a group of points as a whole so that a reference vector aligns with a target vector

    https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d

    Parameters
    ----------
    positions : array
        The positions of the points
    reference : array
        The reference vector for the roratition
    target : array
        The target vector to align with

    Returns
    -------
    positions_rotated : array
        The positions after the rotation
    '''
    # unit vectors
    vec_unit = reference / np.linalg.norm(reference)
    tar_unit = target / np.linalg.norm(target)

    # determine the axis of rotation and angle of rotation
    axis = np.cross(vec_unit, tar_unit)
    sin = np.linalg.norm(axis)
    cos = np.dot(vec_unit, tar_unit)

    # construct the rotation matrix
    skew_matrix = np.array([[0, -axis[2], axis[1]],
                            [axis[2], 0, -axis[0]],
                            [-axis[1], axis[0], 0]])
    rotation_matrix = np.eye(3) + skew_matrix + (1 - cos) / sin ** 2 * np.dot(skew_matrix, skew_matrix)

    return np.array([np.dot(rotation_matrix, p) for p in positions])


def find_clusters(elements, func, show_progress=False):
    '''
    Group elements into clusters

    Parameters
    ----------
    elements : Iterable
    func : func
        This function has two arguments. Both should be element from `elements`.
        Return True if two elements should be grouped into the same cluster.

    Returns
    -------
    clusters : list of list of int
        The clusters formed by elements. Each cluster is represented by a list of the index of elements.
    '''
    n_element = len(elements)
    flag = [0] * n_element
    n_cluster = 0
    clusters = []
    if show_progress:
        from tqdm import tqdm
        _iterator = tqdm(range(n_element))
    else:
        _iterator = range(n_element)
    for i in _iterator:
        if flag[i]:
            continue
        path = {i}
        expand = set()
        n_cluster += 1
        flag[i] = n_cluster
        for j in range(i + 1, n_element):
            if flag[j]:
                continue
            if func(elements[i], elements[j]):
                path.add(j)
                expand.add(j)
                flag[j] = n_cluster
        while expand:
            m = expand.pop()
            for n in range(i + 1, n_element):
                if flag[n]:
                    continue
                if func(elements[m], elements[n]):
                    path.add(n)
                    expand.add(n)
                    flag[n] = n_cluster
        clusters.append(list(path))

    return clusters


def find_clusters_consecutive(elements, func, show_progress=False):
    '''
    Group elements into clusters. If element i and j are in the same group, all elements between i and j will also be put in the same group.

    Parameters
    ----------
    elements : Iterable
    func : func
        This function has two arguments. Both should be element from `elements`.
        Return True if two elements should be grouped into the same cluster.

    Returns
    -------
    clusters : list of list of int
        The clusters formed by elements. Each cluster is represented by a list of the index of elements.
    '''
    n_element = len(elements)
    flag = [0] * n_element
    n_cluster = 0
    clusters = []
    if show_progress:
        from tqdm import tqdm
        _iterator = tqdm(range(n_element))
    else:
        _iterator = range(n_element)
    for i in _iterator:
        if not flag[i]:
            n_cluster += 1
            flag[i] = n_cluster
            clusters.append({i})
        for j in reversed(range(i + 1, n_element)):
            if flag[j]:
                break
            if func(elements[i], elements[j]):
                for n in range(i + 1, j + 1):
                    flag[n] = flag[i]
                    clusters[-1].add(n)
                break

    clusters = [list(c) for c in clusters]

    return clusters



import os
import setuptools

with open('README.md') as f:
    long_description = f.read()

dir_scripts = 'scripts'
scripts = [os.path.join(dir_scripts, f) for f in os.listdir(dir_scripts)]

setuptools.setup(
    name='mstk',
    version='0.3.12',
    author='Zheng Gong',
    author_email='z.gong@outlook.com',
    description='Molecular simulation toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/z-gong/mstk',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        'mstk': ['data/forcefield/*']
    },
    scripts=scripts,
    python_requires='>=3.6',
    # dependency can be a mess if conda and pip are mixed
    # better let user install requirements by themselves
    install_requires=[],
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: OS Independent'
    ]
)



import math
import numpy as np
from mstk.chem.constant import *


class EwaldSum():
    def __init__(self, charges, positions, box, cutoff=1.2, tolerance=5E-4):
        self.charges = np.array(charges)
        self.positions = np.array(positions)
        self.n_atom = len(self.charges)
        self.box = np.array(box)
        self.volume = box[0] * box[1] * box[2]
        self.reciprocal_box = 2 * PI / self.box
        self.cutoff = cutoff
        self.tolerance = tolerance

        self.alpha, self.kmax_x, self.kmax_y, self.kmax_z = self.calc_ewald_parameters()
        self.eikr = self.calc_eikr()  # exp(ikr)
        self.eikr_conj = np.conj(self.eikr)

    def calc_ewald_parameters(self):
        def get_kmax(width, alpha, tolerance, guess):
            def get_ewald_error(kmax):
                temp = kmax * PI / (width * alpha)
                return 0.05 * kmax * math.sqrt(width * alpha) * math.exp(-temp * temp)

            kmax = guess
            error = get_ewald_error(kmax)
            if error < tolerance:
                while error < tolerance and kmax > 0:
                    kmax -= 1
                    error = get_ewald_error(kmax)
                return kmax + 1
            while error > tolerance:
                kmax += 1
                error = get_ewald_error(kmax)
            return kmax

        lx, ly, lz = self.box
        tol = self.tolerance
        alpha = math.sqrt(-math.log(2.0 * tol)) / self.cutoff
        kmaxx = get_kmax(lx, alpha, tol, 10)
        kmaxy = get_kmax(ly, alpha, tol, 10)
        kmaxz = get_kmax(lz, alpha, tol, 10)
        if kmaxx % 2 == 0: kmaxx += 1
        if kmaxy % 2 == 0: kmaxy += 1
        if kmaxz % 2 == 0: kmaxz += 1

        return alpha, kmaxx, kmaxy, kmaxz

    def calc_eikr(self):
        kmax = max(self.kmax_x, self.kmax_y, self.kmax_z)
        array = np.zeros((kmax, self.n_atom, 3), dtype=complex)
        for n in range(self.n_atom):
            for d in range(3):
                array[0][n][d] = complex(1, 0)
                array[1][n][d] = complex(math.cos(self.positions[n][d] * self.reciprocal_box[d]),
                                         math.sin(self.positions[n][d] * self.reciprocal_box[d]))
                for k in range(2, kmax):
                    array[k][n][d] = array[k - 1][n][d] * array[1][n][d]
        return array

    def distance(self, pos1, pos2, cutoff=None):
        delta = pos2 - pos1
        if cutoff is None:
            return delta, np.sqrt(delta.dot(delta))

        delta = delta - np.floor(delta / self.box + 0.5) * self.box
        if np.any(abs(delta) > cutoff):
            return delta, 0
        r = np.sqrt(delta.dot(delta))
        r = 0 if r > cutoff else r
        return delta, r

    def calc_energy_forces(self, ewald=True):
        if ewald:
            e_short, f_short = self.calc_ewald_short()
            e_long, f_long = self.calc_ewald_long()
            e_self = self.calc_ewald_self()
            return e_short + e_long + e_self, f_short + f_long

        energy = 0
        forces = np.zeros((self.n_atom, 3))
        for i in range(self.n_atom):
            for j in range(i + 1, self.n_atom):
                delta, r = self.distance(self.positions[i], self.positions[j])
                e = ONE_4PI_EPS0 * self.charges[i] * self.charges[j] / r
                energy += e
                for d in range(3):
                    dEdR = e / r / r * delta[d]
                    forces[i][d] -= dEdR
                    forces[j][d] += dEdR
        return energy, forces

    def calc_ewald_short(self):
        energy = 0
        forces = np.zeros((self.n_atom, 3))
        for i in range(self.n_atom):
            for j in range(i + 1, self.n_atom):
                delta, r = self.distance(self.positions[i], self.positions[j], self.cutoff)
                if r > 0:
                    e = ONE_4PI_EPS0 * self.charges[i] * self.charges[j] / r
                    alpha_r = self.alpha * r
                    energy += e * math.erfc(alpha_r)
                    for d in range(3):
                        dEdR = e / r / r * delta[d] * math.erfc(alpha_r) \
                               + e * (-2 / PI_SQRT * math.exp(-alpha_r * alpha_r)) \
                               * self.alpha * (-delta[d] / r)
                        forces[i][d] -= dEdR
                        forces[j][d] += dEdR
        return energy, forces

    def calc_ewald_long(self):
        factor_ewald = -1 / (4 * self.alpha ** 2)
        factor_energy = 1 / (2 * self.volume * VACUUM_PERMITTIVITY) \
                        * ELEMENTARY_CHARGE ** 2 / NANO / 1000 * AVOGADRO
        energy = 0
        forces = np.zeros((self.n_atom, 3))
        qxyz = np.zeros(self.n_atom, dtype=complex)
        # for ix in range(-self.kmax_x + 1, self.kmax_x):
        #     for iy in range(-self.kmax_y + 1, self.kmax_y):
        #         for iz in range(-self.kmax_z + 1, self.kmax_z):
        iy_low = 0
        iz_low = 1
        for ix in range(0, self.kmax_x):
            kx = ix * self.reciprocal_box[0]
            for iy in range(iy_low, self.kmax_y):
                ky = iy * self.reciprocal_box[1]
                for iz in range(iz_low, self.kmax_z):
                    kz = iz * self.reciprocal_box[2]
                    S_k = complex(0, 0)
                    for j in range(self.n_atom):
                        eikr_x = self.eikr[ix][j][0] if ix >= 0 else self.eikr_conj[-ix][j][0]
                        eikr_y = self.eikr[iy][j][1] if iy >= 0 else self.eikr_conj[-iy][j][1]
                        eikr_z = self.eikr[iz][j][2] if iz >= 0 else self.eikr_conj[-iz][j][2]
                        qxyz[j] = self.charges[j] * eikr_x * eikr_y * eikr_z
                        S_k += qxyz[j]
                    k2 = kx * kx + ky * ky + kz * kz
                    ak = math.exp(k2 * factor_ewald) / k2
                    energy += ak * (S_k.real ** 2 + S_k.imag ** 2)
                    for j in range(self.n_atom):
                        # force = -1/(2V*eps0)*ak*d(S(k)S(-k))/dRj
                        # f*k=dS(k)/dRj*S(-k)+S(K)*dS(-k)/dRj
                        f = 2 * (qxyz[j].real * S_k.imag - qxyz[j].imag * S_k.real)
                        forces[j][0] -= ak * f * kx
                        forces[j][1] -= ak * f * ky
                        forces[j][2] -= ak * f * kz
                iz_low = 1 - self.kmax_z
            iy_low = 1 - self.kmax_y
        return energy * factor_energy * 2, forces * factor_energy * 2

    def calc_ewald_self(self):
        return - ONE_4PI_EPS0 * (self.charges * self.charges).sum() * self.alpha / PI_SQRT

    @staticmethod
    def create_test(n_atom=100, box=None, seed=0, **kwargs):
        if box is None:
            box = [3.0, 4.0, 5.0]
        np.random.seed(seed)
        charges = np.random.random(n_atom) - 0.5
        charges[-1] = 0 - charges[:-1].sum()
        positions = np.random.random((n_atom, 3)) * 3
        ewald = EwaldSum(charges, positions, box, **kwargs)
        return ewald

    def calc_energy_forces_with_omm(self, ewald=True, short=True, long=True):
        from openmm import openmm as mm, unit
        system = mm.System()
        system.setDefaultPeriodicBoxVectors(*np.diag(self.box))
        nbforce = mm.NonbondedForce()
        system.addForce(nbforce)

        if ewald:
            nbforce.setNonbondedMethod(mm.NonbondedForce.Ewald)
        else:
            nbforce.setNonbondedMethod(mm.NonbondedForce.NoCutoff)
        nbforce.setCutoffDistance(self.cutoff)
        nbforce.setForceGroup(1)
        nbforce.setReciprocalSpaceForceGroup(2)

        for i in range(self.n_atom):
            system.addParticle(0)
            nbforce.addParticle(self.charges[i], 1.0, 0)

        integrator = mm.VerletIntegrator(0.001)
        platform = mm.Platform.getPlatformByName('Reference')
        context = mm.Context(system, integrator, platform)
        context.setPositions(self.positions)
        groups = set()
        if short: groups.add(1)
        if long: groups.add(2)
        state = context.getState(getEnergy=True, getForces=True, groups=groups)
        energy = state.getPotentialEnergy().value_in_unit(unit.kilojoule_per_mole)
        forces = state.getForces(asNumpy=True).value_in_unit(
            unit.kilojoule_per_mole / unit.nanometer)
        return energy, forces



import numpy as np
from mstk.chem.constant import *
from mstk.trajectory import Frame
from mstk.trajectory.handler import TrjHandler


class Dcd(TrjHandler):
    '''
    Read and write step, cell and positions from/to DCD file.

    Currently mstk use chemfiles to support DCD file.
    '''

    def __init__(self, file, mode='r'):
        super().__init__()

        try:
            import chemfiles
        except:
            raise ImportError('Currently mstk use chemfiles to parse DCD format. Cannot import chemfiles')

        if mode not in ('r', 'w', 'a'):
            raise Exception('Invalid mode')

        self._dcd = chemfiles.Trajectory(file, mode, format='DCD')

    def close(self):
        try:
            self._dcd.close()
        except:
            pass

    def get_info(self):
        self.n_frame = self._dcd.nsteps
        if self.n_frame == 0:
            raise Exception('Empty DCD file')
        cf_frame = self._dcd.read()
        self.n_atom = len(cf_frame.atoms)

        return self.n_atom, self.n_frame

    def read_frame(self, i_frame, frame):
        cf_frame = self._dcd.read_step(i_frame)
        cf_cell = cf_frame.cell
        lengths = [i / 10 for i in cf_cell.lengths]
        angles = [i * DEG2RAD for i in cf_cell.angles]
        frame.positions = cf_frame.positions.astype(float) / 10
        frame.cell.set_box([lengths, angles])
        frame.step = frame.step

    def write_frame(self, frame, subset=None, **kwargs):
        '''
        Write a frame into the opened DCD file

        Parameters
        ----------
        frame : Frame
        subset : list of int, optional
        kwargs : dict
            Ignored
        '''
        import chemfiles

        if subset is None:
            positions = frame.positions
        else:
            positions = frame.positions[subset]

        cf_frame = chemfiles.Frame()
        cf_frame.resize(len(positions))
        cf_frame.positions[:] = positions * 10
        cf_frame.cell = chemfiles.UnitCell(frame.cell.lengths * 10, frame.cell.angles * RAD2DEG)
        cf_frame.step = frame.step

        self._dcd.write(cf_frame)


TrjHandler.register_format('.dcd', Dcd)



import itertools
from .system import System
from ..forcefield import *
from ..topology import *
from .. import logger


class NamdExporter():
    '''
    NamdExporter export a :class:`System` to input files for NAMD.

    The following potential functions are currently supported:

    * :class:`~mstk.forcefield.LJ126Term`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.OplsDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.HarmonicImproperTerm`
    * :class:`~mstk.forcefield.DrudeTerm`

    The :class:`~mtools.forcefield.OplsImproperTerm` can be exported, but it will be in the harmonic form.
    Usually it is acceptable if the improper to be described is planer and quite rigid.
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, pdb_out, psf_out, prm_out, **kwargs):
        '''
        Generate input files for NAMD from a system

        Parameters
        ----------
        system : System
        pdb_out : str or None
        psf_out : str or None
        prm_out : str or None
        '''
        if not system.use_pbc:
            raise Exception('PBC required for exporting NAMD')

        if pdb_out is not None:
            NamdExporter._export_pdb(system, pdb_out)
        if psf_out is not None:
            NamdExporter._export_psf(system, psf_out)
        if prm_out is not None:
            NamdExporter._export_prm(system, prm_out)

    @staticmethod
    def _export_pdb(system, pdb_out='conf.pdb'):
        Pdb.save_to(system.topology, pdb_out)

    @staticmethod
    def _export_psf(system, psf_out='top.psf'):
        Psf.save_to(system.topology, psf_out)

    @staticmethod
    def _export_prm(system: System, prm_out='ff.prm'):
        supported_terms = {LJ126Term,
                           HarmonicBondTerm,
                           HarmonicAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm, HarmonicImproperTerm,
                           DrudeTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s' % (', '.join(map(lambda x: x.__name__, unsupported))))

        if OplsImproperTerm in system.ff_classes:
            logger.warning('OplsImproperTerm not supported by NAMD. Exported in harmonic form')

        string = '* Created by mstk\n'
        string += '*\n'

        string += '\nATOMS\n'
        for i, atype in enumerate(system._ff.atom_types.values()):
            # set mass to 0 in case not defined in FF
            string += '  MASS %6i %10s %10.4f\n' % (i + 1, atype.name, max(atype.mass, 0))

        unique_bonds: {(str): BondTerm} = {}
        for bond in system._topology.bonds:
            name = (bond.atom1.type, bond.atom2.type)
            if name not in unique_bonds and tuple(reversed(name)) not in unique_bonds:
                if bond.is_drude:
                    parent = bond.atom2 if bond.atom1.is_drude else bond.atom1
                    pterm = system.polarizable_terms[parent]
                    bterm = HarmonicBondTerm(name[0], name[1], 0.0, pterm.k)
                else:
                    bterm = system.bond_terms[bond]
                unique_bonds[name] = bterm

        unique_angles = {}
        for angle in system._topology.angles:
            name = (angle.atom1.type, angle.atom2.type, angle.atom3.type)
            if name not in unique_angles and tuple(reversed(name)) not in unique_angles:
                unique_angles[name] = system.angle_terms[angle]
        unique_dihedrals = {}
        for dihedral in system._topology.dihedrals:
            name = (dihedral.atom1.type, dihedral.atom2.type,
                    dihedral.atom3.type, dihedral.atom4.type)
            if name not in unique_dihedrals and tuple(reversed(name)) not in unique_dihedrals:
                unique_dihedrals[name] = system.dihedral_terms[dihedral]
        unique_impropers = {}
        for improper in system._topology.impropers:
            name = (improper.atom1.type, improper.atom2.type,
                    improper.atom3.type, improper.atom4.type)
            if name not in unique_impropers and tuple(reversed(name)) not in unique_impropers:
                unique_impropers[name] = system.improper_terms[improper]

        string += '\nBONDS\n'
        string += '''!V(bond) = Kb(b - b0)**2
!Kb: kcal/mole/A**2
!b0: A
!
!      atom type Kb          b0
'''
        for (at1, at2), bterm in unique_bonds.items():
            string += '%10s %10s %16.6f %10.4f\n' % (
                at1, at2, bterm.k / 4.184 / 100, bterm.length * 10)

        string += '\nANGLES\n'
        string += '''!V(angle) = Ktheta(Theta - Theta0)**2
!V(Urey-Bradley) = Kub(S - S0)**2
!Ktheta: kcal/mole/rad**2
!Theta0: degrees
!Kub: kcal/mole/A**2 (Urey-Bradley)
!S0: A
!
!      atom types     Ktheta    Theta0   Kub     S0
'''
        for (at1, at2, at3), aterm in unique_angles.items():
            string += '%10s %10s %10s %12.6f %12.4f\n' % (
                at1, at2, at3, aterm.k / 4.184, aterm.theta * RAD2DEG)

        string += '\nDIHEDRALS\n'
        string += '''!V(dihedral) = Kchi(1 + cos(n(chi) - delta))
!Kchi: kcal/mole
!n: multiplicity
!delta: degrees
!
!atom types             Kchi    n   delta
'''
        for (at1, at2, at3, at4), dterm in unique_dihedrals.items():
            if type(dterm) is OplsDihedralTerm:
                dterm = dterm.to_periodic_term()
            for phase in dterm.phases:
                string += '%10s %10s %10s %10s %12.6f %4i %8.2f\n' % (
                    at1, at2, at3, at4, phase.k / 4.184, phase.n, phase.phi * RAD2DEG)
            if len(dterm.phases) == 0:
                string += '%10s %10s %10s %10s %12.6f %4i %8.2f  ; no dihedral parameters\n' % (
                    at1, at2, at3, at4, 0.0, 1, 0.0)

        string += '\nIMPROPERS\n'
        string += '''!V(improper) = Kpsi(psi - psi0)**2
!Kpsi: kcal/mole/rad**2
!psi0: degrees
!
!      atom types           Kpsi      ignored         psi0
'''
        for (at1, at2, at3, at4), iterm in unique_impropers.items():
            if type(iterm) == HarmonicImproperTerm:
                phi = iterm.phi
            elif type(iterm) == OplsImproperTerm:
                phi = 0.0
            else:
                raise Exception('Only HarmonicImproperTerm and OplsImproperTerm are supported')
            string += '%10s %10s %10s %10s %12.6f %4i %8.2f\n' % (
                at1, at2, at3, at4, iterm.k / 4.184, 0, phi * RAD2DEG)

        string += '\nNONBONDED\n'
        string += '''!V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]
!epsilon: kcal/mole, Eps,i,j = sqrt(eps,i * eps,j)
!Rmin/2: A, Rmin,i,j = Rmin/2,i + Rmin/2,j
!
!      atom  ignored    epsilon      Rmin/2   ignored   eps,1-4       Rmin/2,1-4
'''
        for atype in system._ff.atom_types.values():
            vdw = system._ff.get_vdw_term(atype, atype)
            rmin = vdw.sigma * 2 ** (1 / 6)
            eps14 = vdw.epsilon * system._ff.scale_14_vdw
            string += '%10s %10.4f %12.6f %12.6f %10.4f %12.6f %12.6f  ! %s\n' % (
                atype.name, 0.0, -vdw.epsilon / 4.184, rmin / 2 * 10,
                0.0, -eps14 / 4.184, rmin / 2 * 10, ' '.join(vdw.comments))

        if len(system._ff.pairwise_vdw_terms) != 0:
            string += '\nNBFIX\n'
            string += '!      atom  atom  epsilon      Rmin   eps,1-4       Rmin,1-4\n'
            for atype1, atype2 in itertools.combinations(system._ff.atom_types.values(), 2):
                try:
                    vdw = system._ff.get_vdw_term(atype1, atype2, mixing=False)
                except:
                    continue
                rmin = vdw.sigma * 2 ** (1 / 6)
                eps14 = vdw.epsilon * system._ff.scale_14_vdw
                string += '%10s %10s %12.6f %12.6f %12.6f %12.6f  ! %s\n' % (
                    atype1.name, atype2.name, -vdw.epsilon / 4.184, rmin * 10,
                    -eps14 / 4.184, rmin * 10, ' '.join(vdw.comments))

        string += '\nEND\n'

        with open(prm_out, 'wb') as f:
            f.write(string.encode())



class Residue:
    '''
    A residue is a group of consecutive atoms, e.g. the amino acid in protein, the repeating unit in polymer.

    It can be considered as a special connectivity without energy contribution.
    It is mainly for the convenience of visualization of polymers.
    Small molecule usually contains one single residue.
    '''

    def __init__(self, name='UNK'):
        self.id = -1
        self.id_in_mol = -1
        self.name = name
        self._atoms = []

    def __repr__(self):
        return f'<Residue: {self.name} {self.id}>'

    @property
    def n_atom(self):
        return len(self._atoms)

    @property
    def atoms(self):
        return self._atoms

    def _add_atom(self, atom):
        atom._residue = self
        self._atoms.append(atom)

    def _remove_atom(self, atom):
        atom._residue = None
        self._atoms.remove(atom)



import math
import numpy as np
from pandas import Series

__all__ = [
    'block_average',
    'average_of_blocks',
    'is_converged',
    'efficiency_with_block_size',
    'mean_and_uncertainty',
]


def block_average(series, n_block=5):
    '''
    Calculate the block average, standard deviation and their standard errors

    Parameters
    ----------
    series : Series
        Time series
    n_block : int
        Number of blocks to use

    Returns
    -------
    ave_err_std_err : tuple of tuple of float
        It returns two tuple. The first one is the mean of average of each block and the standard error of this mean.
        The second one is the mean of standard deviation of each block and the standard error of this mean.
    '''
    block_ave_std = average_of_blocks(series, n_block)
    ave_list, std_list = zip(*block_ave_std)
    ave, err_ave = np.mean(ave_list), np.std(ave_list, ddof=1) / math.sqrt(n_block)
    std, err_std = np.mean(std_list), np.std(std_list, ddof=1) / math.sqrt(n_block)
    return (ave, err_ave), (std, err_std)


def average_of_blocks(series, n_block=5):
    '''
    Split data to several blocks and return the average and standard deviation of each block

    Parameters
    ----------
    series : Series
        Time series
    n_block : int
        Number of blocks

    Returns
    -------
    aves : list of tuple of float
        The average and standard deviation of each block
    '''
    array = np.array(series)
    n_points = len(series)
    block_size = n_points // n_block
    blocks = []
    for n in range(n_block - 1):
        blocks.append(array[block_size * n:block_size * (n + 1)])
    blocks.append(array[block_size * (n_block - 1):])
    block_ave_std = [(np.mean(b), np.std(b)) for b in blocks]
    return block_ave_std


def is_converged(series: Series, frac_min=0.5):
    '''
    Determine whether a time series has converged or not

    Parameters
    ----------
    series : Series
        Time series
    frac_min : float
        Consider this time series is converged only if the fraction of converged parts relative to the full series
        is larger than this threshold

    Returns
    -------
    converged : bool
        Converged or not
    when : float
        From when this times series converged
    '''
    from pymbar import timeseries

    n_points = len(series)
    array = np.array(series)
    t0, g, Neff_max = timeseries.detect_equilibration(array, nskip=max(1, n_points // 100))
    if t0 > n_points * (1 - frac_min):
        return False, series.index[t0]
    return True, series.index[t0]


def efficiency_with_block_size(data):
    '''
    Calculate the statistical efficiency of block average with different block size.
    It can be used to determine the maximum number of block size.

    Parameters
    ----------
    data : list of float

    Returns
    -------
    block_size : list of int
        List of block size
    efficiency : list of float
        List of statistic efficiency with different block size
    '''
    array = np.array(data)
    n_points = len(data)
    bsize_list = []
    n_block_list = []
    s_list = []

    for bsize in range(1, int(math.sqrt(n_points))):
        n_block = int(n_points / bsize)
        bsize_list.append(bsize)
        n_block_list.append(n_block)

    for n_block in range(int(math.sqrt(n_points)), 4, -1):
        bsize = int(n_points / n_block)
        bsize_list.append(bsize)
        n_block_list.append(n_block)

    for i, bsize in enumerate(bsize_list):
        n_block = n_block_list[i]
        blocks = np.array_split(array, n_block)
        ave_blocks = [np.mean(block) for block in blocks]
        std_ave_blocks = np.std(ave_blocks, ddof=1)
        s = bsize * std_ave_blocks ** 2 / np.std(array) ** 2
        s_list.append(s)

    return bsize_list, s_list


def mean_and_uncertainty(series: Series, inefficiency=None) -> (float, float):
    from pymbar import timeseries

    ave = np.mean(series)
    array = np.array(series)
    if inefficiency == None:
        inefficiency = timeseries.statistical_inefficiency(array)
    return ave, np.std(array, ddof=1) / math.sqrt(len(array) / inefficiency)



import math
import numpy as np
from mstk.topology import Atom

__all__ = [
    'calc_weighted_average',
    'calc_com',
    'calc_rg',
    'calc_hull_volume',
]


def calc_weighted_average(array, weight):
    '''
    Calculate the weighted average of 2-dimensional data

    Parameters
    ----------
    array : np.array
    weight : np.array
        If weight for all elements equal to zero. Will ignore weight

    Returns
    -------
    average : np.array
    '''
    if all(weight == 0):
        weight = np.ones(weight.shape)
    return np.sum(array * weight[:, np.newaxis], axis=0) / np.sum(weight)


def calc_com(atoms):
    '''
    Calculate the center of mass of a group of atoms

    Parameters
    ----------
    atoms : list of Atom

    Returns
    -------
    com : np.array
    '''
    positions = np.array([atom.position for atom in atoms])
    masses = np.array([atom.mass for atom in atoms])

    return calc_weighted_average(positions, masses)


def calc_rg(atoms):
    '''
    Calculate the radius of gyration of a group of atoms

    Parameters
    ----------
    atoms : list of Atom

    Returns
    -------
    rg : float
    '''
    positions = np.array([atom.position for atom in atoms])
    masses = np.array([atom.mass for atom in atoms])
    if all(masses == 0):
        masses.fill(1.0)

    com = calc_weighted_average(positions, masses)
    positions -= com
    rsq_ew = np.sum(positions ** 2, axis=1)

    return math.sqrt(sum(masses * rsq_ew) / sum(masses))


def calc_hull_volume(atoms):
    '''
    Calculate the volume occupied by a groups of at least four atoms

    Parameters
    ----------
    atoms : list of Atom

    Returns
    -------
    volume : float
    '''
    if len(atoms) < 4:
        raise Exception('At least 4 atoms required for 3D convex hull')

    from scipy.spatial import ConvexHull

    positions = np.array([atom.position for atom in atoms])
    hull = ConvexHull(positions)

    return float(hull.volume)



import os
import subprocess
from subprocess import Popen, PIPE
from pathlib import Path
from mstk import logger
from mstk.errors import SchedulerError
from . import Slurm


class RemoteSlurm(Slurm):
    '''
    Slurm job scheduler running on a remote machine.

    Parameters
    ----------
    queue : str
        The jobs will be submitted to this partition.
    n_proc : int
        The CPU cores a job can use.
    n_gpu : int
        The GPU card a job can use.
    host : str
        The IP address of the remote host that is running Slurm
    username : str
        The username for logging in the remote host
    remote_dir : str
        The default directory to use on the remote host for running calculation
    port : int
        The SSH port for logging in the remote host
    n_node : int
        The nodes a job can use. If 0, then will be decided by slurm.
    exclude: str, Optional
        The nodes to be excluded, in Slurm format
    env_cmd : str, Optional
        The commands for setting up the environment before running real calculations.
        It will be inserted on the top of job scripts.

    Attributes
    ----------
    queue : str
        The jobs will be submitted on this queue.
    n_proc : int
        The CPU cores a job can use.
    n_gpu : int
        The GPU card a job can use.
    n_node : int
        The nodes a job can use. If 0, then will be decided by slurm.
    exclude: str, Optional
        The nodes to be excluded, in Slurm format
    env_cmd : str
        The commands for setting up the environment before running real calculations.
    sh : str
        The default name of the job script
    host : str
        The IP address of the remote host that is running Slurm
    username : str
        The username for logging in the remote host
    remote_dir : str
        The default directory to use on the remote host for running calculation
    port : int
        The SSH port for logging in the remote host
    max_running_hour: int
        The wall time limit for a job in hours.
    cached_jobs_expire : int
        The lifetime of cached jobs in seconds.
    submit_cmd : str
        The command for submitting the job script.
        If is `sbatch` by default. But extra argument can be provided, e.g. `sbatch --qos=debug`.
    '''

    #: Whether or not this is a remote job scheduler
    is_remote = True

    def __init__(self, queue, n_proc, n_gpu, host, username, remote_dir, port=22, n_node=0, exclude=None, env_cmd=None):
        super().__init__(queue=queue, n_proc=n_proc, n_gpu=n_gpu, n_node=n_node, exclude=exclude, env_cmd=env_cmd)

        self.host = host
        self.port = port
        self.username = username
        self.remote_dir = remote_dir

        # be careful about the port option. -p in ssh, -P in scp
        self._ssh_cmd = f'ssh -o ConnectTimeout=5 -o BatchMode=yes -o StrictHostKeyChecking=no -p {self.port} {self.username}@{self.host}'
        self._scp_cmd = f'scp -o ConnectTimeout=5 -o BatchMode=yes -o StrictHostKeyChecking=no -P {self.port}'
        # require rsync 3.1.0 or above
        self._rsync_cmd = f'rsync --human-readable --recursive --links --perms --executability --times --info=progress2 --rsh="ssh -p {self.port}"'

        if not self.is_working():
            raise SchedulerError('Cannot connect to Slurm @ %s' % host)

    def is_working(self) -> bool:
        '''
        Check whether or not Slurm is working normally on the remote machine.

        It calls `sinfo --version` and check the output.

        Returns
        -------
        is : bool
        '''
        logger.info(f'Connecting to Slurm on {self.host}')
        cmd = f'{self._ssh_cmd} sinfo --version'
        sp = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        stdout, stderr = sp.communicate()

        return stdout.decode().startswith('slurm')

    def upload(self, local_dir=None, remote_dir=None):
        '''
        Upload all the files in current local directory to remote directory.

        Parameters
        ----------
        local_dir : dir, optional
            If not set, will use the current dir.
        remote_dir : dir, optional
            If not set, will use the default :attr:`remote_dir`.

        Returns
        -------
        successful : bool
            Whether or not the upload is successful
        '''

        local_dir = local_dir or os.getcwd()
        remote_dir = remote_dir or self.remote_dir

        logger.info(f'Uploading to {remote_dir}')

        mkdir_cmd = f'{self._ssh_cmd} mkdir -p {remote_dir}'
        if subprocess.call(mkdir_cmd, shell=True) != 0:
            return False

        rsync_cmd = f'{self._rsync_cmd} {local_dir}/ {self.username}@{self.host}:{remote_dir}/'
        if subprocess.call(rsync_cmd, shell=True) != 0:
            return False

        return True

    def download(self, remote_dir=None, local_dir=None) -> bool:
        '''
        Upload all the files in remote directory to current local directory.

        Parameters
        ----------
        remote_dir : dir, optional
            If not set, will use the default :attr:`remote_dir`.
        local_dir : dir, optional
            If not set, will use the current dir.

        Returns
        -------
        successful : bool
            Whether or not the download is successful
        '''
        remote_dir = remote_dir or self.remote_dir
        local_dir = local_dir or os.getcwd()

        logger.info(f'Downloading from {remote_dir}')
        rsync_cmd = f'{self._rsync_cmd} {self.username}@{self.host}:{remote_dir}/ {local_dir}/'
        if subprocess.call(rsync_cmd, shell=True) != 0:
            return False

        return True

    def submit(self, sh=None, remote_dir=None):
        '''
        Submit a job script to the Slurm scheduler on the remote machine.

        Parameters
        ----------
        sh : str
            The job script to be submitted.
        remote_dir : str
            The directory to submit the script on the remote machine.

        Returns
        -------
        id : int
        '''
        sh = sh or self.sh
        sh_name = Path(sh).name

        remote_dir_path = Path(remote_dir or self.remote_dir).absolute()
        remote_sh_path = remote_dir_path / sh_name

        logger.info(f'Uploading slurm script to {remote_dir_path}')

        mkdir_cmd = f'{self._ssh_cmd} "mkdir -p {remote_dir_path}"'
        if subprocess.call(mkdir_cmd, shell=True) != 0:
            return False

        scp_cmd = f'{self._scp_cmd} {sh} {self.username}@{self.host}:{remote_sh_path}'
        if subprocess.call(scp_cmd, shell=True) != 0:
            return False

        # Be careful. Use "" for several commands over ssh
        sbatch_cmd = f'{self._ssh_cmd} "cd {remote_dir_path}; {self.submit_cmd} {remote_sh_path}"'
        p = Popen(sbatch_cmd, shell=True, stdout=PIPE, stderr=PIPE)
        b_out, b_err = p.communicate()
        out, err = b_out.strip().decode(), b_err.strip().decode()
        if p.returncode == 0 and out.startswith('Submitted batch job'):
            logger.info(f'{out}')
            return int(out.split()[-1])
        else:
            if out:
                logger.error(f'{out}')
            if err:
                logger.error(f'{err}')
            if not out and not err:
                logger.error(f'sbatch failed')
            return -1

    def kill_job(self, name) -> bool:
        job = self.get_job_from_name(name)
        if job is None:
            return False

        cmd = f'{self._ssh_cmd} scancel {job.id}'
        return subprocess.call(cmd, shell=True) == 0

    def get_all_jobs(self):
        # Show all jobs. Then check the user
        logger.info(f'Querying jobs on remote Slurm...')

        cmd = f'{self._ssh_cmd} scontrol show job'
        sp = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        stdout, stderr = sp.communicate()
        if sp.returncode != 0:
            logger.error(stderr.decode())
            return []

        jobs = []
        for job_str in stdout.decode().split('\n\n'):  # split jobs
            if job_str.startswith('JobId'):
                job = self._get_job_from_str(job_str)
                # Show all jobs. Then check the user
                if job.user == self.username:
                    jobs.append(job)
        return jobs



from mstk.topology import Topology
from mstk.trajectory import Frame
from mstk.trajectory.handler import TrjHandler


class Gro(TrjHandler):
    '''
    Read and write cell, atomic positions and optionally velocities from/to GRO file.
    '''

    def __init__(self, file, mode='r'):
        super().__init__()
        if mode not in ('r', 'w', 'a'):
            raise Exception('Invalid mode')

        if mode == 'r':
            # open it in binary mode so that we can correctly seek despite of line ending
            self._file = open(file, 'rb')
        elif mode == 'a':
            self._file = open(file, 'ab')
        elif mode == 'w':
            self._file = open(file, 'wb')

    def get_info(self):
        try:
            self._file.readline()
            self.n_atom = int(self._file.readline())
        except:
            print('Invalid gro file')
            raise
        self._file.seek(0)

        # read in the file once and build a list of line offsets
        # the last element is the length of whole file
        self._line_offset = [0]
        offset = 0
        for line in self._file:
            offset += len(line)
            self._line_offset.append(offset)
        self._file.seek(0)

        # build a list of frame offsets
        self.n_frame = len(self._line_offset) // (3 + self.n_atom)
        self._frame_offset = []
        for i in range(self.n_frame + 1):
            line_start = (3 + self.n_atom) * i
            self._frame_offset.append(self._line_offset[line_start])

        return self.n_atom, self.n_frame

    def read_frame(self, i_frame, frame):
        # skip to frame i and read only this frame
        self._file.seek(self._frame_offset[i_frame])
        lines = self._file.read(self._frame_offset[i_frame + 1] - self._frame_offset[i_frame]) \
            .decode().splitlines()
        # assume there are velocities. we'll see later
        frame.has_velocity = True
        for i in range(self.n_atom):
            line = lines[i + 2]
            x = float(line[20:28])
            y = float(line[28:36])
            z = float(line[36:44])
            frame.positions[i][:] = x, y, z

            if frame.has_velocity:
                try:
                    vx = float(line[44:52])
                    vy = float(line[52:60])
                    vz = float(line[60:68])
                except:
                    frame.has_velocity = False
                else:
                    frame.velocities[i][:] = vx, vy, vz
        _box = tuple(map(float, lines[self.n_atom + 2].split()))
        if len(_box) == 3:
            frame.cell.set_box(_box)
        elif len(_box) == 9:
            ax, by, cz, ay, az, bx, bz, cx, cy = _box
            frame.cell.set_box([[ax, ay, az], [bx, by, bz], [cx, cy, cz]])
        else:
            raise ValueError('Invalid box')

    def write_frame(self, frame, topology, subset=None, write_velocity=False, **kwargs):
        '''
        Write a frame into the opened GRO file

        Parameters
        ----------
        frame : Frame
        topology : Topology
        subset : list of int, optional
        write_velocity : bool
            Whether or not velocities should be written.
            If set to True but velocities not available in frame, an Exception will be raised.
        kwargs : dict
            Ignored
        '''
        if subset is None:
            subset = list(range(len(frame.positions)))
        if write_velocity and not frame.has_velocity:
            raise Exception('Velocities are requested but not exist in frame')

        string = 'Created by mstk: step= %i, t= %f ps\n' % (frame.step, frame.time)
        string += '%i\n' % len(subset)

        for id in subset:
            atom = topology.atoms[id]
            residue = atom.residue
            pos = frame.positions[id]
            string += '%5i%5s%5s%5i%8.3f%8.3f%8.3f' % (
                (residue.id + 1) % 100000, residue.name[:5], atom.symbol[:5], (atom.id + 1) % 100000,
                pos[0], pos[1], pos[2])
            if write_velocity:
                vel = frame.velocities[id]
                string += '%8.4f%8.4f%8.4f' % (vel[0], vel[1], vel[2])
            string += '\n'

        a, b, c = frame.cell.vectors
        string += ' %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n' % (
            a[0], b[1], c[2], a[1], a[2], b[0], b[2], c[0], c[1])

        self._file.write(string.encode())
        self._file.flush()

TrjHandler.register_format('.gro', Gro)



import numpy as np
from .atom import Atom

__all__ = [
    'VirtualSite',
    'TwoLineSite',
    'ThreePlaneSite',
    'TIP4PSite',
]


class VirtualSite:
    '''
    Base class for virtual site definitions.

    This class should not be constructed directly. Use its subclasses instead or call :create:.
    '''

    _class_map = {}

    def __init__(self):
        self.parents = []
        self.parameters = []

    def calc_position(self):
        '''
        Calculate the position of virtual site from parent atoms.

        Returns
        -------
        position : array_like
        '''
        raise NotImplementedError('Method not implemented')

    @staticmethod
    def register(klass):
        '''
        Register a virtual site class so that it can be created.

        Parameters
        ----------
        klass : subclass of VirtualSite
        '''
        VirtualSite._class_map[klass.__name__] = klass

    @staticmethod
    def create(type, parents, parameters):
        '''
        Factory function for creating a new virtual site

        Parameters
        ----------
        type: str
            The type of virtual site to be created
        parents : list of Atom
            The parent atoms of the new virtual site
        parameters : list of float
            Parameters for calculating the position of virtual site from parent atoms

        Returns
        -------
        virtual_site : subclass of VirtualSite
        '''

        try:
            cls = VirtualSite._class_map[type]
        except:
            raise Exception('Unknown virtual site type. Valid types: ' + ', '.join(VirtualSite._class_map.keys()))

        return cls(parents, parameters)


class TwoLineSite(VirtualSite):
    '''
    A virtual site in the same line defined by two atoms

    Parameters
    ----------
    parents : list of Atom
    parameters : list of float
    '''

    def __init__(self, parents, parameters):
        if len(parents) != 2 or len(parameters) != 1:
            raise Exception('Invalid number of parents or parameters')

        super().__init__()
        self.parents = parents[:]
        self.parameters = parameters[:]

    def calc_position(self):
        if not all(atom.has_position for atom in self.parents):
            raise Exception('Position for parent atoms are required')

        a1, a2 = self.parents
        p1 = self.parameters[0]
        return a1.position * p1 + a2.position * (1 - p1)


VirtualSite.register(TwoLineSite)


class ThreePlaneSite(VirtualSite):
    '''
    A virtual site in the same plane defined by three atoms

    Parameters
    ----------
    parents : list of Atom
    parameters : list of float
    '''

    def __init__(self, parents, parameters):
        if len(parents) != 3 or len(parameters) != 2:
            raise Exception('Invalid number of parents or parameters')

        super().__init__()
        self.parents = parents[:]
        self.parameters = parameters[:]

    def calc_position(self):
        if not all(atom.has_position for atom in self.parents):
            raise Exception('Position for parent atoms are required')

        a1, a2, a3 = self.parents
        p1, p2 = self.parameters
        return a1.position * p1 + a2.position * p2 + a3.position * (1 - p1 - p2)


VirtualSite.register(ThreePlaneSite)


class TIP4PSite(VirtualSite):
    '''
    A virtual site used in TIP4P model. This is a special case of ThreePlaneSite.

    Parameters
    ----------
    parents : list of Atom
    parameters : list of float
    '''

    def __init__(self, parents, parameters):
        if len(parents) != 3 or len(parameters) != 1:
            raise Exception('Invalid number of parents or parameters')

        super().__init__()
        self.parents = parents[:]
        self.parameters = parameters[:]

    def calc_position(self):
        if not all(atom.has_position for atom in self.parents):
            raise Exception('Position for parent atoms are required')

        a_O, a_H1, a_H2 = self.parents
        d = self.parameters[0]
        vec = a_H1.position + a_H2.position - 2 * a_O.position
        vec_unit = vec / np.sqrt(vec.dot(vec))
        return a_O.position + d * vec_unit


VirtualSite.register(TIP4PSite)



from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk.chem.element import Element


class XyzTopology():
    '''
    Generate Topology from XYZ file.

    XYZ format only records the type (or atomic symbol) and position of atoms.
    There is no real topology, so all atoms are assumed to be in the same molecule.
    The first column is treated as atom type instead of name or symbol.

    Parameters
    ----------
    file : str
    kwargs : dict
        Ignored

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> xyz = XyzTopology('input.xyz')
    >>> topology = xyz.topology

    >>> XyzTopology.save_to(topology, 'output.xyz')
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._parse(file)

    def _parse(self, file):
        with open(file) as f:
            n_atom = int(f.readline().strip())
            mol = Molecule()
            try:
                mol.name = f.readline().strip().split()[0]
            except:
                pass
            for i in range(n_atom):
                words = f.readline().split()
                atom = Atom()
                atom.type = words[0]
                element = Element.guess_from_atom_type(atom.type)
                atom.symbol = element.symbol
                atom.mass = element.mass
                atom.name = atom.symbol + str(mol.n_atom + 1)
                atom.position = tuple(map(lambda x: float(x) / 10, words[1:4]))
                mol.add_atom(atom)

        self.topology.update_molecules([mol])

    @staticmethod
    def save_to(top, file, **kwargs):
        '''
        Save topology into a XYZ file.

        The name of the first molecule in the topology will be written as the remark of the topology.
        Only atom type and positions are written.
        If atom type is empty, use atom symbol instead.

        Parameters
        ----------
        top : Topology
        file : str
        '''
        if not top.has_position:
            raise Exception('Position is required for writing XYZ file')

        string = '%i\n' % top.n_atom
        string += '%s\n' % top.molecules[0].name
        for atom in top.atoms:
            pos = atom.position * 10
            string += '%-8s %11.5f %11.5f %11.5f\n' % (
                atom.type or atom.symbol, pos[0], pos[1], pos[2])
        with open(file, 'wb')as f:
            f.write(string.encode())


Topology.register_format('.xyz', XyzTopology)



import itertools
from mstk.forcefield.forcefield import ForceField
from mstk.forcefield.ffterm import *
from mstk.forcefield.errors import *
from mstk import logger


class Padua:
    '''
    Generate ForceField from the force field file of `fftool`.

    The `fftool` developed by Agilio Pauda use OPLS convention with extension of Drude polarization,
    which contains following terms:

    * :class:`~mstk.forcefield.AtomType` with equivalent tables, masses and partial charges
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.OplsImproperTerm`
    * :class:`~mstk.forcefield.DrudeTerm`

    The energy terms are in k/2 form for bond, angle, dihedral and improper.
    The length is in unit of Angstrom, and angle is in unit of degree.
    The energies for bond, angle and vdW are in unit of kJ/mol/A^2, kJ/mol/rad^2 and kJ/mol.
    The two values for LJ interaction are sigma and epsilon.

    A force field term describing one topology element should only appear once.
    If there are duplicated terms, an Exception will be raised.

    Parameters
    ----------
    file : str

    Attributes
    ----------
    forcefield : ForceField
    '''

    def __init__(self, file):
        ff = ForceField()
        ff.lj_mixing_rule = ForceField.LJ_MIXING_GEOMETRIC
        ff.vdw_long_range = ForceField.VDW_LONGRANGE_CORRECT
        ff.vdw_cutoff = 1.2  # nm
        ff.scale_14_vdw = 0.5
        ff.scale_14_coulomb = 0.5

        self._parse(ff, file)
        self._setup(ff)

        self.forcefield = ff

    def _parse(self, ff, file):
        with open(file) as f:
            lines = f.read().splitlines()

        _section = ''
        for line in lines:
            if line.startswith('#') or line.strip() == '':
                continue

            if line.lower().startswith('atom'):
                _section = 'atoms'
                continue
            elif line.lower().startswith('bond'):
                _section = 'bonds'
                continue
            elif line.lower().startswith('angl'):
                _section = 'angles'
                continue
            elif line.lower().startswith('dihe'):
                _section = 'dihedrals'
                continue
            elif line.lower().startswith('impro'):
                _section = 'impropers'
                continue
            elif line.lower().startswith('polar'):
                _section = 'polarizations'
                continue

            words = line.strip().split()
            if _section == 'atoms':
                self._parse_atom(ff, words)
            elif _section == 'bonds':
                self._parse_bond(ff, words)
            elif _section == 'angles':
                self._parse_angle(ff, words)
            elif _section == 'dihedrals':
                self._parse_dihedral(ff, words)
            elif _section == 'impropers':
                self._parse_improper(ff, words)
            elif _section == 'polarizations':
                self._parse_polarization(ff, words)
            else:
                raise Exception('Invalid header for fftool: %s' % line)

    def _setup(self, ff):
        # If this is a polarizable FF, add an AtomType and VdwTerm for Drude particles
        if ff.is_polarizable:
            type_drude = 'DP_'
            if type_drude not in ff.atom_types:
                dtype = AtomType(type_drude)
                vdw = LJ126Term(type_drude, type_drude, 0.0, 0.0)
                ff.add_term(dtype)
                ff.add_term(vdw)

        # If H* is defined in polarizable term, alpha of H will be merged into attached heavy atoms and the term will be removed
        if 'H*' in ff.polarizable_terms:
            hterm = ff.polarizable_terms.pop('H*')
            for pterm in ff.polarizable_terms.values():
                pterm.merge_alpha_H = hterm.alpha
            logger.info(f'H* found in polarizable term. '
                        f'Polarizability of H will be merged into attached heavy atoms')

    def _parse_atom(self, ff, words):
        atype = AtomType(words[0])
        atype.mass = float(words[2])
        atype.charge = float(words[3])
        (atype.eqt_bond, atype.eqt_ang_c, atype.eqt_ang_s,
         atype.eqt_dih_c, atype.eqt_dih_s,
         atype.eqt_imp_c, atype.eqt_imp_s) = [words[1]] * 7
        if words[4] == 'lj':
            sigma = float(words[5]) / 10  # convert from A to nm
            epsilon = float(words[6])
            lj = LJ126Term(atype.name, atype.name, epsilon, sigma)
            ff.vdw_terms[lj.name] = lj
        else:
            raise Exception('Unsupported vdw potential form: %s' % (words[4]))

        if atype.name not in ff.atom_types.keys():
            ff.atom_types[atype.name] = atype
        else:
            raise Exception('Duplicated atom type: %s' % str(atype))

    def _parse_bond(self, ff, words):
        if words[2] not in ['harm', 'cons']:
            raise Exception('Unsupported bond function: %s' % (words[2]))
        term = HarmonicBondTerm(words[0], words[1], length=float(words[3]) / 10,
                                k=float(words[4]) * 100 / 2, fixed=(words[2] == 'cons'))
        if term.name in ff.bond_terms.keys():
            raise Exception('Duplicated bond term: %s' % str(term))
        ff.bond_terms[term.name] = term

    def _parse_angle(self, ff, words):
        if words[3] not in ['harm', 'cons']:
            raise Exception('Unsupported angle function: %s' % (words[3]))
        term = HarmonicAngleTerm(words[0], words[1], words[2], theta=float(words[4]) * DEG2RAD,
                                 k=float(words[5]) / 2, fixed=(words[3] == 'cons'))
        if term.name in ff.angle_terms.keys():
            raise Exception('Duplicated angle term: %s' % str(term))
        ff.angle_terms[term.name] = term

    def _parse_dihedral(self, ff, words):
        if words[4] not in ['opls']:
            raise Exception('Unsupported dihedral function: %s' % (words[4]))
        term = PeriodicDihedralTerm(words[0], words[1], words[2], words[3])
        k1, k2, k3, k4 = list(map(lambda x: float(x) / 2, words[5:9]))
        if k1 != 0: term.add_phase(0.0, k1, 1)
        if k2 != 0: term.add_phase(PI, k2, 2)
        if k3 != 0: term.add_phase(0.0, k3, 3)
        if k4 != 0: term.add_phase(PI, k4, 4)
        if term.name in ff.dihedral_terms.keys():
            raise Exception('Duplicated dihedral term: %s' % str(term))
        ff.dihedral_terms[term.name] = term

    def _parse_improper(self, ff, words):
        '''
        For improper i-j-k-l
        mstk always use CHARMM convention, i is the central atom
        fftool use OPLS convention, k is the central atom
        '''
        if words[4] not in ['opls']:
            raise Exception('Unsupported improper function: %s' % (words[4]))
        if {float(words[5]), float(words[7]), float(words[8])} != {0}:
            raise Exception('k1, k3 and k4 for opls improper should equal to 0')
        term = OplsImproperTerm(words[2], words[0], words[1], words[3], k=float(words[6]) / 2)
        if term.name in ff.improper_terms.keys():
            raise Exception('Duplicated improper term: %s' % str(term))
        ff.improper_terms[term.name] = term

    def _parse_polarization(slef, ff, words):
        name, mass, charge, k, alpha, thole = words
        term = DrudeTerm(name, float(alpha) / 1000, float(thole))
        term.mass = float(mass)
        term.k = float(k) / 2 * 100  # convert from kJ/mol/A^2 to kJ/mol/nm^2
        if term.name in ff.polarizable_terms.keys():
            raise Exception('Duplicated drude term: %s' % str(term))
        ff.polarizable_terms[term.name] = term


class Monomer:
    def __init__(self, name: str, charge: float, dipole: float, alpha: float = None):
        self.name = name
        self.charge = charge
        self.dipole = dipole
        self.alpha = alpha or -1  # set to -1 if not provided
        self.atoms: [str] = []  # list of atom types

    def __repr__(self):
        return '<Monomer: %s>' % self.name


class Dimer:
    C0 = 0.25
    C1 = 0.11

    def __init__(self, m1: Monomer, m2: Monomer, distance: float, scale_factor: float = None):
        '''
        If scale factor is None, then need to calculate scale factor later
        based on charge, dipole, polarizability and distance of monomers
        '''
        self.monomer1 = m1
        self.monomer2 = m2
        self.distance = distance
        self.scale_factor = scale_factor

    def __repr__(self):
        return '<Dimer: %s %s>' % (self.monomer1.name, self.monomer2.name)

    def predict_scale_epsilon(self):
        m1, m2 = self.monomer1, self.monomer2
        if (m1.charge != 0 or m1.dipole != 0) and m1.alpha == -1:
            raise Exception(
                'Error in predicting scaling factor: alpha required for charged or polar fragment: %s' % m1.name)
        if (m2.charge != 0 or m2.dipole != 0) and m2.alpha == -1:
            raise Exception(
                'Error in predicting scaling factor: alpha required for charged or polar fragment: %s' % m2.name)

        k = self.C0 * self.distance ** 2 * (m1.charge ** 2 / m1.alpha + m2.charge ** 2 / m2.alpha) + \
            self.C1 * (m1.dipole ** 2 / m1.alpha + m2.dipole ** 2 / m2.alpha)
        self.scale_factor = round(1 / (1 + k), 3)


class PaduaLJScaler:
    '''
    PaduaLJScaler implements the empirical LJ scaling scheme of Goloviznina Kateryna.

    A scaling file should be providing,
    which provides information about fragments, dimers, atom types belongs to each fragments,
    charges and dipoles of each fragments and distances of dimers.

    Scaling factor can also be provided directly in DIMERS section.
    K_SAPT2 section is ignored.

    After the scaling file been loaded, all the LJ126Terms in a ForceField can be scaled by calling
    function :func:`scale`.

    Examples
    --------
    >>> scaler = PaduaLJScaler('ljscale.txt')
    >>> ff = ForceField.open('input.ff')
    >>> scaler.scale(ff)

    Parameters
    ----------
    file : str

    Attributes
    ----------
    scale_sigma : float
        Scaling factor for sigma, which equals to 1.0 by default.
        Other value can be provided by the input file by adding a line `SCALE_SIGMA value`

    '''

    def __init__(self, file):
        self.scale_sigma = 1.0
        self._monomers = []
        self._dimers = []
        self._parse(file)

    def _parse(self, file):
        with open(file) as f:
            lines = f.read().splitlines()

        _section = ''
        for line in lines:
            if line.strip() == '' or line.startswith('#'):
                continue

            words = line.strip().split()
            if words[0] == 'MONOMERS':
                _section = 'MONOMERS'
                continue
            elif words[0] == 'DIMERS':
                _section = 'DIMERS'
                continue
            elif words[0] == 'K_SAPT2':
                _section = 'K_SAPT2'
                continue
            elif words[0] == 'ATOMS':
                _section = 'ATOMS'
                continue
            elif words[0] == 'SCALE_SIGMA':
                self.scale_sigma = float(words[1])
                continue

            if _section == 'MONOMERS':
                name = words[0].strip('+').strip('-')
                alpha = None
                if len(words) > 3:
                    alpha = float(words[3])
                self._monomers.append(Monomer(name, float(words[1]), float(words[2]), alpha))

            elif _section == 'DIMERS':
                m1 = next(m for m in self._monomers if m.name == words[0])
                m2 = next(m for m in self._monomers if m.name == words[1])
                scale = None
                if len(words) > 3:
                    try:
                        scale = float(words[3])
                    except:
                        pass
                self._dimers.append(Dimer(m1, m2, float(words[2]), scale))

            elif _section == 'ATOMS':
                monomer = next(m for m in self._monomers if m.name == words[0])
                monomer.atoms = words[1:]

    def scale(self, ff):
        '''
        Scale epsilon and sigma of all the LJ126Terms in a ForceField.

        Both the self-vdW terms and pairwise-vdW terms are scaled, as long as it is LJ126Term.

        Parameters
        ----------
        ff : ForceField

        Returns
        -------
        all_scaled : bool
            Whether or not all the vdW terms are been scaled.
            If it's False, it means fragment or dimer data not found for some LJ126Terms.
        '''
        _all_scaled = True
        for type1, type2 in itertools.combinations(ff.atom_types.values(), 2):
            # must scale pairwise vdW terms first
            # because they may be generated from self vdW terms by combination rule
            # the scaled pairwise terms must be added into the ff set
            try:
                vdw = ff.get_vdw_term(type1, type2)
            except FFTermNotFoundError:
                continue
            if type(vdw) is LJ126Term:
                if vdw.epsilon == 0:
                    continue
                if not self.scale_lj(vdw):
                    _all_scaled = False
                ff.add_term(vdw, replace=True)
        for vdw in ff.vdw_terms.values():
            if type(vdw) == LJ126Term:
                if vdw.epsilon == 0:
                    continue
                if not self.scale_lj(vdw):
                    _all_scaled = False

        return _all_scaled

    def scale_lj(self, term):
        '''
        Scale the epsilon and sigma of a LJ126Term.

        If scaling factor for epsilon is successfully predicted,
        then epsilon will be scaled by this factor, and sigma will be scaled by predefined scaling factor.
        If scaling factor for epsilon is not predicted (because of fragment or dimer not found),
        then both epsilon and sigma will be untouched.

        Parameters
        ----------
        term : LJ126Term

        Returns
        -------
        scaled : bool
        '''
        k_eps = self.predict_scale_epsilon(term.type1, term.type2)
        if k_eps is None:
            return False

        term.epsilon *= k_eps
        term.comments.append('eps*%.3f' % k_eps)
        k_sig = self.scale_sigma
        if k_sig != 1.0:
            term.sigma *= k_sig
            term.comments.append('sig*%.3f' % k_sig)

        return True

    def predict_scale_epsilon(self, at1, at2):
        '''
        Predict the scaling factor for epsilon of the LJ126Term between two atom types
        with empirical formulation.

        Parameters
        ----------
        at1 : str
            The first atom type
        at2 : str
            The second atom type

        Returns
        -------
        k : float or None
            float value means successful prediction.
            None means fragment or dimer data not found.
        '''
        m1 = m2 = None
        for monomer in self._monomers:
            if at1 in monomer.atoms:
                m1 = monomer
            if at2 in monomer.atoms:
                m2 = monomer
        if m1 is None or m2 is None:
            return None

        for dimer in self._dimers:
            if (m1 == dimer.monomer1 and m2 == dimer.monomer2) or (
                    m2 == dimer.monomer1 and m1 == dimer.monomer2):
                dimer = dimer
                break
        else:
            return None

        if dimer.scale_factor is None:
            dimer.predict_scale_epsilon()

        return dimer.scale_factor


ForceField.registor_format('.ff', Padua)
